#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DESCRIPTION

""" Beta function and beta class indicators. """


# DEPENDENCIES

import tirl
from tirl.tirlobject import InstanceCounterMeta


# DEFINITIONS

__version__ = tirl.__version__


# IMPLEMENTATION

def beta_function(func):
    """
    Decorator that indicates when a certain function/method is untested. Warns
    the user that results from the respective function/method might be
    incorrect.

    :param func: untested function or method
    :type func: Callable

    :returns: untested function or method with a warning message
    :rtype: Callable

    """
    import warnings
    from functools import wraps
    @wraps(func)
    def wrapped(*args, **kwargs):
        warnings.warn(
            f"{func.__name__} is a beta function in TIRL {__version__}. "
            f"Use with caution: results might be incorrect.")
        return func(*args, **kwargs)
    return wrapped


class BetaClassMeta(InstanceCounterMeta):
    """
    Metaclass that indicates when a certain class is untested. Warns the user
    that results from the respective class might be incorrect.

    """
    def __new__(cls, name, bases, attrs):
        import warnings
        from functools import wraps
        initfunc = attrs['__init__']
        @wraps(initfunc)
        def wrapped(*args, **kwargs):
            self = args[0]
            warnings.warn(f"{self.__class__.__name__} is a beta class in TIRL "
                f"{__version__}. Use with caution: results might be incorrect.")
            initfunc(*args, **kwargs)
        attrs['__init__'] = wrapped
        return super(BetaClassMeta, cls).__new__(cls, name, bases, attrs)
