#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np


# TIRL IMPORTS

from tirl import settings as ts
from tirl.costs.cost import Cost


# DEFINITIONS

from tirl.constants import *


# IMPLEMENTATION

class CostMSD(Cost):
    """
    CostMSD - Mean Squared Differences cost function term

    """
    def __init__(self, source, target, target_filter=None,
                 source_prefilter=None, source_postfilter=None,
                 maskmode="and", matchmode="l2_norm", normalise=True,
                 logger=None, **metaparameters):
        """
        Initialisation of CostMSD.

        :param source:
            Moving image. Values of the source image are interpolated at the
            points of the target domain.
        :type source: TImage
        :param target:
            Fixed image. Values of the source image are interpolated at the
            points of the target domain.
        :type target: TImage
        :param target_filter:
            Filtering that is applied to the target TImage at the construction
            of the Cost object. The target filter must be an Operator,
            otherwise it must be callable with a single input argument for the
            target TImage, and the callable must return a TImage with the
            Domain of the target TImage (not a copy). If None (default), no
            filtering is performed.
        :type target_filter: Operator or Callable or None
        :param source_prefilter:
            Filtering that is applied to the source TImage at the construction
            of the Cost object. The source prefilter must be an Operator,
            otherwise it must be callable with a single input argument for the
            source TImage, and the callable must return a TImage with the
            Domain of the source TImage (not a copy). If None (default), no
            filtering is performed.
        :type source_prefilter: Operator or Callable or None
        :param source_postfilter:
            Filtering that is applied to the source TImage after interpolation,
            before the cost is calculated. The source postfilter must be an
            Operator, otherwise it must be callable with a single input
            argument for the resampled source TImage, and the callable must
            return a TImage with the Domain of the resampled source TImage
            (i.e. the domain of the target TImage; not a copy). If None
            (default), no filtering is performed.
        :type source_postfilter: Operator or Callable or None
        :param maskmode:
            Defines how the masks of the source and the target TImage should be
            combined into one mask. The following methods are supported:
            1. "and": combined mask values are calculated as the geometric mean
                      of corresponding source and target mask values (default).
                      For binary masks, this is equivalent to the AND operation.
            2. "or": combined mask values are calculated as the quadratic mean
                     of corresponding source and target mask values. For binary
                     masks this is equivalent to the OR operation.
            3. callable: function(target_mask, source_mask) -> combined_mask.
        :type maskmode: str or Callable
        :param matchmode:
            If the filtered target image and the prefiltered, resampled and
            post-filtered source image do not have conformant tensor shapes,
            this routine will be applied automatically to create conformable
            shapes. The following modes are available:
            1. None: turned off: raise error if tensors are non-conformant
            2. "broadcast": broadcast tensor shapes according to NumPy rules
            3. "l1_norm" (default): reduce to scalar using the L1-norm
            4. "l2_norm": reduce to scalar using the L2-norm
            5. callable: function(target, source_on_target) -> trg, src_on_trg
        :type matchmode: str or Callable or None
        :param normalise:
            If True (default), the cost is normalised by the number of non-zero
            voxels in the combined mask, or the number of voxels in the target
            image if no mask was set.
        :type normalise: bool
        :param metaparameters:
            Additional keyword arguments.
        :type metaparameters: Any

        """
        # Call superclass initialisation
        super(CostMSD, self).__init__(
            source, target, target_filter=target_filter,
            source_prefilter=source_prefilter,
            source_postfilter=source_postfilter, maskmode=maskmode,
            matchmode=matchmode, logger=logger, **metaparameters)

        # Set class-specific metaparameters
        if isinstance(normalise, bool):
            self.metaparameters.update(normalise=normalise)
        else:
            raise TypeError(f"Expected bool value for normalise option, "
                            f"got {normalise.__class__.__name__}")

    def function(self, target, source, combined_mask):
        """
        Calculates the scalar cost based on squared differences between the
        target and the source constituents. If the normalise option is set
        (default), the cost is normalised by the number of foreground voxels.

        """
        if combined_mask is not None:
            nonzero = ~np.isclose(combined_mask, 0)
        else:
            nonzero = np.full(shape=target.domain.numel, fill_value=True)

        if not np.any(nonzero):
            raise RuntimeError("All elements of the combined mask are zero.")

        # Restrict calculation to mask and enforce TENSOR_MAJOR order without
        # changing the layout of the input images. This enables broadcasting.
        stensors = source.voxels[nonzero]
        if source.order != TENSOR_MAJOR:
            stensors = np.moveaxis(stensors, -1, 0)
        ttensors = target.voxels[nonzero]
        if target.order != TENSOR_MAJOR:
            ttensors = np.moveaxis(ttensors, -1, 0)
        # Adjust types to the operation
        costs = (stensors.astype(ts.DEFAULT_FLOAT_TYPE) -
                 ttensors.astype(ts.DEFAULT_FLOAT_TYPE)) ** 2
        if combined_mask is not None:
            costs = costs * combined_mask[nonzero]
        if self.metaparameters.get("normalise"):
            return np.sum(costs) / stensors.shape[-1]
        else:
            return np.sum(costs)

    def jacobian_aos(self, dim=None):
        """
        Returns the voxelwise total derivative of the cost with respect to the
        local displacement vector. The result is a 2D array:
        (voxels, displacement dimension).

        Note that the cost Jacobian is calculated on the voxel grid of the
        target image. As a consequence, to obtain valid parameter Jacobians of
        the total cost, the result of this function must be multiplied by
        transformation jacobians mapped into the voxel space of the target
        image.

        """
        source = self.source  # only evaluate once
        mask = self.combine_masks(self.target.mask, source.mask)
        diff = source - self.target
        diff.apply_mask(mask, normalise=True)
        diff.order = VOXEL_MAJOR
        source.order = VOXEL_MAJOR
        taxes = diff.taxes

        if dim is None:
            dim = source.vaxes
            grad = np.stack(np.gradient(source, axis=dim), axis=-1)
            n = grad.ndim
            jac = np.einsum(diff, [Ellipsis, *taxes], grad,
                            [Ellipsis, *taxes, n])
        elif isinstance(dim, int):
            dim = source.vaxes[dim]
            grad = np.gradient(source, axis=dim)
            jac = np.einsum(diff, [Ellipsis, *taxes], grad, [Ellipsis, *taxes])
        else:
            raise TypeError("Invalid dimension specification.")

        jac /= self.target.tsize
        return jac.reshape((source.domain.numel, -1))

    def dx(self, dim=None):
        """
        Returns the voxelwise total derivative of the source with respect to the
        local displacement vector. The result has the following shape:
        (voxels, *tensor_shape, gradient_directions).

        Note that the cost Jacobian is calculated on the voxel grid of the
        target image. As a consequence, to obtain valid parameter Jacobians of
        the total cost, the result of this function must be multiplied by
        transformation jacobians mapped into the voxel space of the target
        image.

        """
        source = self.source  # only evaluate once
        mask = self.combine_masks(self.target.mask, source.mask)
        # diff = source - self.target
        # diff.order = VOXEL_MAJOR
        source.order = VOXEL_MAJOR
        # taxes = diff.taxes

        if dim is None:
            dim = sorted(source.vaxes)
            grad = np.stack(np.gradient(source.data, axis=dim), axis=-1)
            # Neumann boundary condition
            # for ax in dim:
            #     slicer = (slice(None),) * ax + (0, Ellipsis)
            #     grad[slicer] = 0
            #     slicer = (slice(None),) * ax + (-1, Ellipsis)
            #     grad[slicer] = 0
            grad = grad.reshape((source.domain.numel, *source.tshape, len(dim)))
        elif isinstance(dim, int):
            dim = source.vaxes[dim]
            grad = np.gradient(source, axis=dim)
            # Neumann boundary condition
            # slicer = (slice(None),) * dim + (0, Ellipsis)
            # grad[slicer] = 0
            # slicer = (slice(None),) * dim + (-1, Ellipsis)
            # grad[slicer] = 0
            grad = grad.reshape((source.domain.numel, *source.tshape))

        else:
            raise TypeError("Invalid dimension specification.")

        # Apply mask to the gradients
        if mask is not None:
            mask = mask.ravel()
            grad = (grad.T * (mask / np.max(mask))).T

        return grad
