#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np


# TIRL IMPORTS

from tirl import settings as ts
from tirl.interpolators.interpolator import Interpolator


# IMPLEMENTATION

class ScipyInterpolator(Interpolator):
    """
    ScipyInterpolator class - wrapper around Scipy.ndimage.map_coordinates

    """

    def interpolate(self, coordinates, input_array=None, **ipkwargs):

        from scipy.ndimage.interpolation import map_coordinates

        kw = {k: v for k, v in self.kwargs.get("ipkwargs", dict()).items()}
        # Override the ones that were specified for the call
        kw.update(ipkwargs)

        if isinstance(coordinates, (tuple, list)):
            coordinates = np.asarray(coordinates).reshape((1, -1))
        try:
            res = map_coordinates(input_array, coordinates.T, **kw)
        except Exception:
            print("Interpolation failed with the following input: ",
                  input_array.shape, coordinates.shape, kw)
            raise
        return res


class ScipyRegularGridInterpolator(Interpolator):

    def __init__(self, source, tensor_axes=(), radius=1, hold=False,
                 threads=-1, verbose=False, ipkwargs=None, **kwargs):
        """
        Initialisation of ScipyRegularGridInterpolator

        """
        # Set default arguments for the RegularGridInterpolator.
        # These parameters will repress out-of-bounds errors when values need
        # to be interpolated outside the domain of known values.
        default_ipkwargs = dict(bounds_error=False, fill_value=ts.FILL_VALUE)
        if isinstance(ipkwargs, dict):
            default_ipkwargs.update(ipkwargs)
        ipkwargs = default_ipkwargs

        # Call parent-class initialisation
        super(ScipyRegularGridInterpolator, self).__init__(
            source, tensor_axes=tensor_axes, radius=radius, hold=hold,
            threads=threads, verbose=verbose, ipkwargs=ipkwargs, **kwargs)


    def interpolate(self, coordinates, input_array=None, **ipkwargs):
        """
        Interpolation routine.

        """
        from scipy.interpolate import RegularGridInterpolator

        kw = {k: v for k, v in self.kwargs.get("ipkwargs", dict()).items()}
        # Override the ones that were specified for the call
        kw.update(ipkwargs)

        if isinstance(coordinates, (tuple, list)):
            coordinates = np.asarray(coordinates).reshape((1, -1))
        try:
            grid = tuple(np.arange(dim) for dim in input_array.shape)
            ip = RegularGridInterpolator(grid, input_array, **kw)
            res = ip(coordinates)
        except:
            print("Interpolation failed with the following input: ",
                  input_array.shape, coordinates.shape, kw)
            raise
        else:
            return res


class ScipyNearestNeighbours(Interpolator):

    def __init__(self, source, tensor_axes=(), radius=1, hold=False,
                 threads=-1, verbose=False, ipkwargs=None, **kwargs):
        """
        Initialisation of ScipyRegularGridInterpolator

        """
        # Set default arguments for the RegularGridInterpolator.
        # These parameters will repress out-of-bounds errors when values need
        # to be interpolated outside the domain of known values.
        default_ipkwargs = dict(rescale=False)
        if isinstance(ipkwargs, dict):
            default_ipkwargs.update(ipkwargs)
        ipkwargs = default_ipkwargs

        # Call parent-class initialisation
        super(ScipyNearestNeighbours, self).__init__(
            source, tensor_axes=tensor_axes, radius=radius, hold=hold,
            threads=threads, verbose=verbose, ipkwargs=ipkwargs, **kwargs)

    def interpolate(self, coordinates, input_array=None, **ipkwargs):

        from scipy.interpolate import NearestNDInterpolator

        kw = {k: v for k, v in self.kwargs.get("ipkwargs", dict()).items()}
        # Override the ones that were specified for the call
        kw.update(ipkwargs)

        if isinstance(coordinates, (tuple, list)):
            coordinates = np.asarray(coordinates).reshape((1, -1))
        try:
            vcoords = np.stack(
                np.meshgrid(*tuple(range(dim) for dim in input_array.shape),
                            indexing="ij"), axis=-1)
            ip = NearestNDInterpolator(vcoords.reshape((-1, input_array.ndim)),
                input_array.ravel(), **kw)
            res = ip(coordinates)
        except:
            print("Interpolation failed with the following input: ",
                  input_array.shape, coordinates.shape, kw)
            raise
        else:
            return res
