#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

from attrdict import AttrMap
from scipy import optimize
from scipy.optimize import minimize, Bounds


# TIRL IMPORTS

from tirl.optimisation.optimiser import Optimiser


# IMPLEMENTATION

class SPOpt(Optimiser):
    """
    SPOpt class - TIRL optimiser class that wraps scipy.optimize.minimize().

    """

    def __init__(self, target, *cost_items, method="Powell", jac="2-point",
                 hess="2-point", hessp=None, constraints=(), tol=None,
                 callback=None, options=None, logger=None, **metaparameters):
        """
        Initialisation of SPOpt.

        :param target:
            Transformation or OptimisationGroup object whose parameters will
            be optimised. During the optimisation, the parameters of this
            object will be updated in place.
        :type target:
            Transformation or OptimisationGroup
        :param cost_items:
            One or more Cost or Regulariser objects that together define the
            objective function for the optimisation.
        :type cost_items: Cost or Regulariser
        :param metaparameters:
            Additional keyword argument to the Optimiser object.
        :type metaparameters: Any

        Bounds specifications are disabled for __init__(). These should be set
        and updated directly within the Transformation or OptimisationGroup.

        """
        super(SPOpt, self).__init__(
            target, *cost_items, logger=logger, **metaparameters)

        self.metaparameters.update(
            method=method,
            jac=jac,
            hess=hess,
            hessp=hessp,
            constraints=constraints,
            tol=tol,
            callback=callback,
            options=options,
        )

    def __call__(self, *args, **kwargs):
        """
        Perform optimisation by calling scipy.optimize.minimize.

        """
        p = AttrMap(self.metaparameters)

        # Define parameter bounds using SciPy's Bounds object
        lb, ub = self.transformation.get_bounds().T
        bounds = Bounds(lb, ub)

        # Introducing a level of indirection because scipy.minimize requires
        # parameters in ndarray format, and the objective function must take
        # the parameter vector as its first argument.
        parameters = self.transformation.get()

        def visualise(x):
            self.log(x)
            src = self.costs[0].source
            trg = self.costs[0].target
            (src - trg).preview()

        def printparams(x):
            print(x)

        def fprime(x, *args):
            return optimize.optimize.approx_fprime(
                x, self.measure_cost, 0.00001)

        def fpp(x, p, *args):
            return optimize.optimize.approx_fhess_p\
                (x, p, fprime, 0.00001, *args)

        method = self.metaparameters.get("method")
        jac = self.metaparameters.get("jac")
        hess = self.metaparameters.get("hess")
        hessp = self.metaparameters.get("hessp")
        if method.lower() in ("newton-cg", "trust-ncg", "trust-krylov"):
            if (jac is None) or (jac == "2-point"):
                self.metaparameters.update(jac=fprime)
            if hess in ("2-point", "3-point") and (hessp is None):
                self.metaparameters.update(hess=None)
                self.metaparameters.update(hessp=fpp)

        # Perform the optimisation
        opt = minimize(self.measure_cost, parameters, method=p.method,
                       jac=p.jac, hess=p.hess, hessp=p.hessp, bounds=bounds,
                       constraints=p.constraints, tol=p.tol,
                       callback=printparams, options=p.options)

        # Set the optimum parameters
        self.transformation.set(opt.x)

    def measure_cost(self, x, *args):
        """
        Sets the parameters and returns the current total cost. This method
        constitutes an indirection from the total_cost method, and is needed
        for compatibility with the SciPy optimisers.

        """
        self.transformation.set(x)
        return self.total_cost()
