#!/usr/bin/env python

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np

# TIRL IMPORTS

from tirl.domain import Domain


# DEFINITIONS

from tirl.constants import *


# IMPLEMENTATION


class Point(Domain):
    """
    Point implements a special case of Domain, when the extent is a single
    point. It is in essence a non-compact Domain, but it is a special subclass
    for added convenience features.

    """

    def __new__(cls, coordinates, *transformations, name=None):
        """
        Construction of Point.

        """
        # Return existing Point without copying
        if isinstance(coordinates, Point):
            return coordinates

        # Create new Point (as non-compact Domain)
        coordinates = np.atleast_2d(coordinates)
        if len(coordinates) != 1:
            raise AssertionError("Only one set of coordinates should be given "
                                 "to initialise Point!")
        obj = super(Point, cls).__new__(
            cls, coordinates, *transformations, offset=None,
            name=name, storage=MEM, n_threads=1)
        return obj

    def __init__(self, coordinates, *transformations, name=None):
        """
        Initialisation of Point.

        """
        if isinstance(coordinates, Point):
            pass
        else:
            coordinates = np.atleast_2d(coordinates)
            super(Point, self).__init__(
                coordinates, *transformations, offset=None,
                name=name, storage=MEM)

    @property
    def x(self):
        return self.get_intermediate_coordinates().ravel()

    @property
    def y(self):
        return self.get_physical_coordinates().ravel()

    def __bool__(self):
        """
        The truth value of a Point is True, if it is different from
        the origin. This attribute has a special meaning for rotations.

        """
        return not np.allclose(self.y, 0)

    def copy(self):
        """
        Copy constructor.

        """
        return type(self)(self.x, *self.chain, name=self.name)

    def _dump(self):
        """
        Creates TIRLObject dump.

        """
        # Call TIRLObject's _dump method
        objdump = super(Domain, self)._dump()
        objdump.update({
            "coordinates": self.x.tolist(),
            "chain": self.chain,
            "name": self.name
        })
        return objdump

    @classmethod
    def _load(cls, dump):
        """
        Creates a Point object from a TIRLObject dump.

        """
        coordinates = dump.pop("coordinates")
        chain = dump.pop("chain")
        name = dump.pop("name")
        return cls(coordinates, *chain, name=name)


if __name__ == "__main__":
    print("This module is not intended for direct execution!")
