#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEVELOPMENT NOTES

"""
28-June-2020 Istva N Huszar:
This class now Works with v2.1.
23-June-2020 Istvan N Huszar:
This is a legacy class from v1.0, which currently does not work with v2.0,
although it is a high-priority task to fix this class.

"""


# DEPENDENCIES

import numpy as np


# TIRL IMPORTS

from tirl.regularisers.txregulariser import TransformationRegulariser


# IMPLEMENTATION

class RegBendingEnergy(TransformationRegulariser):
    """
    RegBendingEnergy - membrane energy regularisation. Calculates the
    regularisation cost as the sum of the second derivative of all
    displacements on a dense domain.

    """
    RESERVED_KWARGS = ("transformation", "weight", "logger", "metaparameters")

    def __init__(self, transformation, weight=1.0, logger=None,
                 **metaparameters):
        """
        Initialisation of RegBendingEnergy.

        :param transformation:
            Displacement field type transformation.
        :type transformation: TxDisplacementField
        :param weight: regularisation weight (default=1.0)
        :type weight: Union[int, float, np.integer, np.floating]
        :param logger:
            If None (default), logs will be redirected to the RootLogger
            instance. You may alternatively specify a different Logger instance
            or its name.
        :type logger: None or str or Logger

        """
        super(RegBendingEnergy, self).__init__(
            transformation=transformation, weight=weight,
            logger=logger, **metaparameters)

    @property
    def transformation(self):
        return self.metaparameters.get("transformation")

    @transformation.setter
    def transformation(self, tx):
        if not hasattr(tx, "domain") or not hasattr(tx, "field"):
            raise TypeError(
                "Diffusion regularisation can only work with a displacement "
                "field type transformation.")
        else:
            self.metaparameters.update(transformation=tx)

    def function(self):
        if self.transformation.domain.iscompact:
            field = self.transformation.std().absolute().field()
        else:
            from tirl.domain import Domain
            d0 = self.transformation.domain
            dense_domain = Domain(
                self.transformation.dense_shape, *d0.chain,
                offset=d0.offset[1:])
            field = self.transformation.std().absolute().field(dense_domain)

        grad1 = np.stack(np.gradient(field.data, axis=field.vaxes), axis=0)
        grad2 = np.stack(np.gradient(grad1, axis=field.vaxes), axis=0)
        res = np.sum(np.abs(grad2)) / field.domain.numel
        self.log("Bending energy: {}".format(res * self.weight))
        return res

    def jacobian(self, parameters):
        return NotImplementedError()

    def hessian(self, parameters):
        return NotImplementedError()
