#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


import unittest
import numpy as np

import tirl
from tirl.constants import *
from tirl.timage import TImage
from tirl.costs.mind import CostMIND
from tirl.transformations.linear.rotation import TxRotation2D


def resources(*args):
    return tirl.sharedir("resources", *args)


class TestConstruction(unittest.TestCase):

    def setUp(self) -> None:
        self.orig = tirl.testimg()
        self.orig.dtype = "<f4"
        self.orig.centralise()
        self.rot = TxRotation2D(20).apply(self.orig.copy())

    def test_construction(self):
        cost = CostMIND(self.rot, self.orig)
        # cost.mind(self.orig).preview()
        self.assertIsInstance(cost, CostMIND)


class TestScalar(unittest.TestCase):

    def setUp(self) -> None:
        self.orig = tirl.testimg()
        self.orig.dtype = "<f4"
        self.orig.centralise()
        self.rot = TxRotation2D(20).apply(self.orig.copy())

    def test_scalar_cost(self):
        cost = CostMIND(self.rot, self.orig)
        # cost.mind(self.orig).preview()
        self.assertIsInstance(cost, CostMIND)
        self.assertGreater(cost(), 0)

    def test_scalar_cost_masked(self):
        cost = CostMIND(self.rot, self.orig)
        fullcost = cost()
        self.orig.mask = TImage(
            resources("testimage", "testimg_edge_mask.tif"))
        cost = CostMIND(self.rot, self.orig)
        self.assertNotEqual(fullcost, cost())


class TestGradient(unittest.TestCase):

    def setUp(self) -> None:
        self.orig = tirl.testimg()
        self.orig.dtype = "<f4"
        self.orig.centralise()
        self.rot = TxRotation2D(20).apply(self.orig.copy())

    def test_cost_gradient(self):
        cost = CostMIND(self.rot, self.orig, kernel=MK_STAR)
        jac = cost.dx()
        n_voxels = self.orig.vsize
        n_neighbours = 4       # 4-point STAR stencil
        n_layers = 3           # RGB
        ndim = self.orig.vdim  # number of partial first derivatives
        tshape = (n_neighbours, n_layers)
        self.assertTupleEqual(jac.shape, (n_voxels, *tshape, ndim))
        self.assertTrue(np.all(np.isfinite(jac)))

    def test_cost_gradient_masked(self):
        cost = CostMIND(self.rot, self.orig)
        fulljac = cost.dx()
        self.assertTrue(np.all(np.isfinite(fulljac)))
        self.orig.mask = TImage(
            resources("testimage", "testimg_edge_mask.tif"))
        cost = CostMIND(self.rot, self.orig)
        jac = cost.dx()
        self.assertFalse(np.allclose(jac, fulljac))
        self.assertTrue(np.all(np.isfinite(jac)))


if __name__ == '__main__':
    unittest.main()
