#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


import os
import unittest
import numpy as np

import tirl.fsl
from tirl.timage import TImage

# DEFINITIONS

FSL_FNIRT_DISPLACEMENT_FIELD       = 2006
FSL_CUBIC_SPLINE_COEFFICIENTS      = 2007
FSL_DCT_COEFFICIENTS               = 2008
FSL_QUADRATIC_SPLINE_COEFFICIENTS  = 2009

def resources(*args):
    return tirl.sharedir("resources", *args)


class TestFLIRTMapping(unittest.TestCase):

    def test_mapping(self):
        # Load images
        hrfile = resources("fsl", "T1_biascorr.nii.gz")
        highres = TImage(hrfile, sform=False, dtype="<f4")
        stdfile = resources("fsl", "MNI152_T1_2mm.nii.gz")
        standard = TImage(stdfile, sform=False, dtype="<f4")
        expfile = os.path.join(
            resources("fsl", "T1_to_MNI_lin_noresampblur.nii.gz"))
        expectation = TImage(expfile, dtype="<f4")

        # Load and set up transformation
        matfile = resources("fsl", "T1_to_MNI_lin.mat")
        flirttx = tirl.fsl.load_mat(matfile, input=hrfile, reference=stdfile)
        standard.domain.chain = flirttx.inverse()
        registered = highres.evaluate(standard.domain)
        registered.domain.chain = registered.domain.chain[:-len(flirttx)]
        # (registered - expectation).preview()
        self.assertTrue(np.isclose(
            np.median(registered.data - expectation.data), 0, atol=1e-3))
        # registered.domain.chain = expectation.domain.chain
        # registered.snapshot("reg_lin.nii.gz")


class TestFNIRTMapping(unittest.TestCase):

    def test_mapping(self):
        # Load images
        hrfile = resources("fsl", "T1_biascorr.nii.gz")
        highres = TImage(hrfile, sform=False, dtype="<f4")
        stdfile = resources("fsl", "MNI152_T1_2mm.nii.gz")
        standard = TImage(stdfile, sform=False, dtype="<f4")
        expfile = resources("fsl", "T1_to_MNI_nonlin.nii.gz")
        expectation = TImage(expfile, dtype="<f4")

        # Load and set up transformation
        warp = resources("fsl", "T1_to_MNI_nonlin_field.nii.gz")
        warp = tirl.fsl.load_warp(warp, moving=hrfile)
        standard.domain.chain = warp
        registered = highres.evaluate(standard.domain)
        registered.domain.chain = registered.domain.chain[:-len(warp)]
        # (registered - expectation).preview()
        self.assertTrue(np.isclose(
            np.median(registered.data - expectation.data), 0, atol=1e-3))
        # registered.domain.chain = expectation.domain.chain
        # registered.snapshot("reg.nii.gz")


class TestFNIRTOtherFormats(unittest.TestCase):

    @unittest.expectedFailure
    def test_cubic_spline_coefficient_field(self):
        self.fail()

    @unittest.expectedFailure
    def test_dct_coefficient_field(self):
        self.fail()

    @unittest.expectedFailure
    def test_quadratic_spline_coefficient_field(self):
        self.fail()

    @unittest.expectedFailure
    def test_unknown_field(self):
        self.fail()

    @unittest.expectedFailure
    def test_invalid_field(self):
        self.fail()


if __name__ == '__main__':
    unittest.main()
