#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


import unittest
import numpy as np

from tirl.transformations.linear.shear import TxShear


class MyTestCase(unittest.TestCase):

    def test_mat2indexedparams(self):
        # Identity matrix
        identity = np.asarray([[1, 0, 0],
                               [0, 1, 0],
                               [0, 0, 1]])
        params, indices = TxShear._mat2params(identity)
        self.assertTrue(np.allclose(params, np.asarray([0, 0, 0])))
        self.assertTupleEqual(tuple(indices), (1, 2, 5))

        # Example matrix
        example = np.asarray([[1, -0.15, 0],
                              [0, 1, 0.23],
                              [0.001, 0, 1]])
        params, indices = TxShear._mat2params(example)
        self.assertTrue(np.allclose(params, np.asarray([-0.15, 0.001, 0.23])))
        self.assertTupleEqual(tuple(indices), (1, 6, 5))

    def test2d_construction_upper_tri_params(self):
        tx = TxShear(2, -0.5)
        self.assertIsInstance(tx, TxShear)
        self.assertTupleEqual(tuple(tx.parameters), (-0.5,))

        self.assertEqual(tx.dim, 2)
        indices = tx.metaparameters.get("indices")
        self.assertEqual(indices, 1)
        # parammask = tx.metaparameters.get("parameter_mask")
        # expected = np.zeros((2, 2), dtype=np.bool)
        # expected.flat[indices] = True
        # self.assertTrue(np.all(parammask == expected))

    def test2d_construction_lower_tri_params(self):
        tx = TxShear(2, lower=(-0.5,))
        self.assertIsInstance(tx, TxShear)
        self.assertTupleEqual(tuple(tx.parameters), (-0.5,))

        self.assertEqual(tx.dim, 2)
        indices = tx.metaparameters.get("indices")
        self.assertEqual(indices, 2)
        # parammask = tx.metaparameters.get("parameter_mask")
        # expected = np.zeros((2, 2), dtype=np.bool)
        # expected.flat[indices] = True
        # self.assertTrue(np.all(parammask == expected))

    def test3d_construction_upper_tri_params(self):
        tx = TxShear(3, -0.1, 0.2, 0.0051)
        self.assertIsInstance(tx, TxShear)
        self.assertTrue(np.allclose(tx.parameters, [-0.1, 0.2, 0.0051]))

        self.assertEqual(tx.dim, 3)
        indices = tx.metaparameters.get("indices")
        self.assertTupleEqual(tuple(indices), (1, 2, 5))
        # parammask = tx.metaparameters.get("parameter_mask")
        # expected = np.zeros((3, 3), dtype=np.bool)
        # expected.flat[indices] = True
        # self.assertTrue(np.all(parammask == expected))

    def test3d_construction_lower_tri_params(self):
        tx = TxShear(3, lower=(-0.1, 0.2, 0.0051))
        self.assertIsInstance(tx, TxShear)
        self.assertTrue(np.allclose(tx.parameters, (-0.1, 0.2, 0.0051)))
        self.assertEqual(tx.dim, 3)
        indices = tx.metaparameters.get("indices")
        self.assertTupleEqual(tuple(indices), (3, 6, 7))
        # parammask = tx.metaparameters.get("parameter_mask")
        # expected = np.zeros((3, 3), dtype=np.bool)
        # expected.flat[indices] = True
        # self.assertTrue(np.all(parammask == expected))


if __name__ == '__main__':
    unittest.main()
